﻿using HeuristicLab.Common;
using HeuristicLab.Core;
using HeuristicLab.Encodings.SymbolicExpressionTreeEncoding;
using HeuristicLab.Persistence.Default.CompositeSerializers.Storable;

namespace HeuristicLab.Problems.DataAnalysis.Symbolic.TreeDistance {
  [StorableClass]
  [Item("Isomorphic tree distance calculator", "An operator that computes the distance between two symbolic expression trees based on the size of their largest common subtree.")]
  public class IsomorphicTreeDistanceCalculator : Item, IDistanceCalculator {
    private readonly SymbolicExpressionTreeNodeSimilarityComparer comparer;

    public IsomorphicTreeDistanceCalculator() {
      comparer = new SymbolicExpressionTreeNodeSimilarityComparer {
        MatchConstantValues = true,
        MatchVariableNames = true,
        MatchVariableWeights = true
      };
    }

    protected IsomorphicTreeDistanceCalculator(IsomorphicTreeDistanceCalculator original, Cloner cloner)
      : base(original, cloner) {
      this.comparer = original.comparer;
    }

    public override IDeepCloneable Clone(Cloner cloner) {
      return new IsomorphicTreeDistanceCalculator(this, cloner);
    }

    public double CalculateDistance(ISymbolicExpressionTree t1, ISymbolicExpressionTree t2) {
      return 1 - MaxCommonSubtreeSimilarity(t1, t2, comparer); // we return the complement of the similarity to resemble the notion of a "distance"
    }

    public static double MaxCommonSubtreeSimilarity(ISymbolicExpressionTree a, ISymbolicExpressionTree b, SymbolicExpressionTreeNodeSimilarityComparer comparer) {
      int max = 0;
      var rootA = a.Root.GetSubtree(0).GetSubtree(0);
      var rootB = b.Root.GetSubtree(0).GetSubtree(0);
      foreach (var aa in rootA.IterateNodesBreadth()) {
        int lenA = aa.GetLength();
        if (lenA <= max) continue;
        foreach (var bb in rootB.IterateNodesBreadth()) {
          int lenB = bb.GetLength();
          if (lenB <= max) continue;
          int matches = SymbolicExpressionTreeMatching.Match(aa, bb, comparer);
          if (max < matches) max = matches;
        }
      }
      return 2.0 * max / (rootA.GetLength() + rootB.GetLength());
    }
  }
}
